#
# Copyright (C) 2023 LLCZ00
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.  
#
"""
omniserver.servers:
    - Server and RequestHandler classes
    - DNS MixIns for RequestHandler and Client classes,
        - Extends classes to provide DNS-specific functionality

Main inheritable classes:
    TCPHandler
    UDPHandler
    DNSHandler
    DNSHandlerTCP
    HTTPHandler

TODO:

"""    
import socketserver
import socket
import threading
import sys
import os
from time import sleep
from dnslib import DNSRecord,RR,A,RCODE,QTYPE
from http.server import SimpleHTTPRequestHandler

from omniserver.version import __version__

__all__ = ["TCPHandler", "UDPHandler", "DNSHandlerTCP",
            "DNSHandler", "HTTPHandler"]

"""Base Classes
- Inherited by RequestHandler and Server subclasses in omniserver.servers to provide overridable data handling methods
"""

class ThreadedBaseServer(socketserver.ThreadingMixIn, socketserver.BaseServer):
    """Base class for TCP and UDP server classes
    """
    allow_reuse_address = True
    proto = "Generic"
    verbose=True
    def __init__(self, server_address, RequestHandlerClass):
        super().__init__(server_address, RequestHandlerClass)
        self.active = False
        
    def __str__(self):
        return f"{self.proto} Server {self.server_address[0]}:{self.server_address[1]}"
        
    def activate(self):
        """Bind socket and activate server
        """
        if not self.active: # Don't call bind or server_activate if they've already been called (error)
            try:
                self.server_bind()
                self.server_activate() # Enable listening
            except:
                self.server_close()
                print(f"{self} Binding/activation failed")
                raise
            else:
                self.active = True
                if self.verbose:
                    print(f"{self} started...")
                    
    def serve_forever(self, poll_interval=0.5):
        """Overriding socketserver method to ensure socket
        is activated
        """
        if not self.active:
            self.activate()
        try:
            super().serve_forever(poll_interval)
        except KeyboardInterrupt: # CTRL-C to exit without raising errors
            self.shutdown()
        
    def handle_request(self):
        """Handle one request, overriding socketserver
        method to ensure socket is activated
        """
        if not self.active:
            self.activate()
        super().handle_request()
 

class BaseRequestHandler:
    """Base class for RequestHandler classes, basically an extension 
    of socketserver.BaseRequestHandler
    
    Sub classes must override send() and recv() with their protocol-specific implimentation.
    The send/recv and send_data/recv_data methods are seperated so as to easily allow 
    for bypassing the incoming/outgoing filter methods, if necessary.
    """
    keepalive = True
    def __init__(self, request, client_address, server):
        self.request = request
        self.client_address = client_address
        self.server = server
        self.proto = server.proto if hasattr(server, "proto") else ""
        self.verbose = server.verbose if hasattr(server, "verbose") else True
        
        self.setup()
        try:
            self.handle()
        finally:
            self.finish()
            
    def setup(self):
        """Called before handle
        """
        pass
        
    def handle(self):
        """Called by socketserver upon connection/request
        
        Keeps connection alive with client if 'keepalive' is
        True, otherwise connection is closed between each request.
        Overriding is not necessary, unless you don't want to use
        the typical 'recv, send' communication pattern.
        """        
        try:
            while True:
                data = self.recv_data()
                if not data:
                    break   
                response = self.response(data)
                self.send_data(response)
                if not self.keepalive:
                    break
        except Exception as e:
            print(f"Exception occured within {self.proto} server")
            if self.verbose:
                print(e)
        
    def finish(self):
        """Called after handle
        """
        pass
            
    def send(self, data):
        """Override with protocol/use specific send method
        
        Use directly to bypass outgoing()
        """
        raise NotImplementedError
        
    def outgoing(self, resp): # Encode data to be sent
        """Process outgoing response/data before it's sent
        
        Default: encode data and add newline
        """
        if type(resp) is not bytes:
            resp = f"{resp}\n".encode("utf-8")
        return resp
        
    def send_data(self, resp):
        """Main method to be used for sending data
        
        Sends data after it's passed through outgoing()
        """
        resp = self.outgoing(resp)
        if self.verbose:
            print(f"[{self.client_address[0]}:{self.client_address[1]}] {self.proto} data sent.")
        self.send(resp)
        
    def recv(self):
        """Override with protocol/use specific recv method
        
        Use directly to bypass incoming()
        """
        raise NotImplementedError
        
    def incoming(self, data): # Decode recieved data from bytes
        """Process incoming data as it's recieved
        
        Default: Decode and strip incoming data
        """
        try:
            data = data.decode()
        except:
            pass
        return data.strip()
    
    def recv_data(self):
        """Main method to be used for recieving data
        
        Returns data from recv(), after being passed through incoming()
        """
        data = self.incoming(self.recv())
        if data:
            if self.verbose:
                print(f"[{self.client_address[0]}:{self.client_address[1]}] {self.proto} data recieved: {data}")
            return data
        
    def response(self, data):
        """Recieves data processed by incoming(), returns 
        response data to be processed by outgoing()
        
        Called by handle()
        """
        return f"Default {self.proto} response"


class DNSRequestHandler(BaseRequestHandler):
    """RequestHandler class to extend BaseRequestHandler with DNS-specific methods
    
    Declares self.records and self.default_ip
    """
    keepalive=False
    def __init__(self, request, client_address, server):
        self.default_ip = server.default_ip if hasattr(server, "default_ip") else None
        self.records = server.records if hasattr(server, "records") else None
        super().__init__(request, client_address, server)
  
    def incoming(self, data):
        """
        Parses recieved data and returns dnslib.DNSRecord object
        """
        try:
            data = DNSRecord.parse(data)
        except:
            return None
        return data
        
    def outgoing(self, resp):
        """
        Ensure DNS response is packed
        """
        if type(resp) is not bytearray:
            resp = resp.pack()
        return resp

    def response(self, request):
        """
        Passes request to resolve and returns response
        - Main overridable method
        - Returns dnslib.DNSRecord object
        """
        reply = request.reply()
        if self.records:
            reply = self.resolve_from_zone(reply, self.records)            
        if not reply.rr:
            if self.default_ip and (reply.q.qtype == QTYPE.A or reply.q.qtype == QTYPE.AAAA):
                reply.add_answer(RR(str(reply.q.qname), QTYPE.A, rdata=A(self.default_ip), ttl=60))            
            else: # Add other record types here
                reply.header.rcode = RCODE.NXDOMAIN # "Could not be resolved"       
        return reply
        
    def resolve_from_zone(self, request, records): # Maybe fold into Resolver class
        """
        Resolve DNS query using given list of DNS records
        - Return reply with added answers
        """
        qname = request.q.qname
        qtype = QTYPE[request.q.qtype]
        zone = [(rr.rname,QTYPE[rr.rtype],rr) for rr in records]
        for name,rtype,rr in zone:
            if getattr(qname,'__eq__')(name) and (qtype == rtype or qtype == 'ANY' or rtype == 'CNAME'):
                request.add_answer(rr)       
                if rtype in ['CNAME','NS','MX','PTR']:
                    for a_name,a_rtype,a_rr in zone:
                        if a_name == rr.rdata.label and a_rtype in ['A','AAAA']:
                            request.add_ar(a_rr)
        return request 
    
    # Overriding these for logging purposes    
    def recv_data(self):
        """Main method to be used for recieving data
        
        Returns data from recv(), after being passed through incoming()
        """
        return self.incoming(self.recv())
        
    def send_data(self, resp):
        """Main method to be used for sending data
        
        Sends data after it's passed through outgoing()
        """
        if resp.rr and self.verbose:
           print(f"[{self.client_address[0]}:{self.client_address[1]}] Query/response: {QTYPE[resp.q.qtype]} {resp.q.qname} -> {resp.get_a().rdata}")  
        self.send(self.outgoing(resp))

""" RequestHandler and Server Classes
- RequestHandler classes can be subclasses and passed to Server class
- Servers accept RequestHandlers upon initialization
- All servers threading capable
"""

# UDP #

class TCPServer(ThreadedBaseServer, socketserver.TCPServer):
    """Threaded TCP server class, TLS/SSL compatible
    
    Listens for and accepts requests from remote TCP clients,
    passes requests to given RequestHandler class.
    """
    proto = "TCP"
    def __init__(self, server_address, RequestHandlerClass):
        """Constructor.  May be extended, do not override."""
        ThreadedBaseServer.__init__(self, server_address, RequestHandlerClass)
        self.socket = socket.socket(self.address_family, self.socket_type)
                
    def enable_ssl(self, context):
        """Enable TLS/SSL on server socket
        
        :param context: SSLContext to wrap socket with (required)
        :rtype: bool
        """
        try:
            self.socket = context.wrap_socket(self.socket, server_side=True)
        except Exception as e:
            print("TLS/SSL failed.")
            raise e
        self.proto = f"{self.proto} (SSL)"
        return True


class TCPHandler(BaseRequestHandler):
    """Handler class for TCP server connections/requests
    
    Overridable methods:
    - incoming
    - outgoing
    - response
    """
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Called by send_data(), after it passes the data through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        self.request.sendall(data)       
        
    def recv(self, buffer: int = 1024):
        """Protocol specific method for recieving data to remote socket
        
        Called by recv_data()
        
        :param buffer: Size of recieve buffer
        :type buffer: int
        :returns data: Data recieved, after being processed through self.incoming()
        """
        return self.request.recv(buffer)


# UDP #

class UDPServer(ThreadedBaseServer, socketserver.UDPServer):
    """Threaded UDP server class 
    
    Extends the functionality of socketserver.UDPServer.
    Passes requests to given RequestHandler class.
    """
    proto = "UDP"
    def __init__(self, server_address, RequestHandlerClass):
        """Constructor.  May be extended, do not override."""
        ThreadedBaseServer.__init__(self, server_address, RequestHandlerClass)
        self.socket = socket.socket(self.address_family, self.socket_type)


class UDPHandler(BaseRequestHandler):
    """Handler class for UDP server requests
    
    Overridable methods:
    - incoming
    - outgoing
    - response
    """
    keepalive = False
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Called by send_data(), after it passes the data through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        self.request[1].sendto(data, self.client_address)
        
    def recv(self):
        """Protocol specific method for recieving data to remote socket
        
        Called by recv_data(), which passes the returned data through self.incoming()
        
        :returns data: Data recieved from remote socket
        """
        return self.request[0].strip()


# DNS #

class DNSServerTCP(TCPServer):
    """ Threaded DNS Server (TCP)
    
    Subclass of omniserver.servers.TCPServer that declares self.default_ip and self.record,
    which are needed for resolving incoming queries.Inherits DNS-specific functions from 
    omniserver.servers.BaseDNSServer, general server functions from omniserver.servers.TCPServer
    """
    proto = "DNS (TCP)"
    def __init__(self, server_address, RequestHandlerClass, default_ip=None):
        super().__init__(server_address, RequestHandlerClass)
        self.default_ip = default_ip
        self.records = []  
         
    def add_record(self, zone_record):
        """Add zone-style record to compare incoming queries against
        
        Ex: "google.com 60 IN A 192.168.100.1"
        
        :param zone_record: DNS zone record string to add to server's 'record'
        :type zone_record: str
        """
        self.records.append(*RR.fromZone(zone_record))
        
    def add_zonefile(self, zonefile):
        """Add all DNS records from given DNS zone file
        
        :param zonefile: Path to file containing zone records
        :type zonefile: str
        """
        with open(zonefile) as file:
            rrs = RR.fromZone(file)
            for rr in rrs:
                self.records.append(rr)


class DNSHandlerTCP(DNSRequestHandler):
    """Handler class for DNS server requests (TCP)
     
    Must be used to exchange information larger than 512 bytes.
    DNSRequestMixIn overrides incoming(), outgoing(), and response(). 
    response() can be further overriden to customize DNS resolution method. 
    self.records and self.default_ip inherited from DNSBaseHandler
    """
    def recv(self):
        """Protocol specific method for recieving data to remote socket
        
        Called by recv_data(), which passes the returned data through self.incoming()
        
        :returns data: Data recieved from remote socket
        """
        data = self.request.recv(8192).strip()
        sz = struct.unpack('>H', data[:2])[0]
        if sz < len(data) - 2:
            raise Exception("Wrong size of TCP packet")
        elif sz > len(data) - 2:
            raise Exception("Too big TCP packet")
        return data[2:]

    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Called by send_data(), after it passes the data through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        sz = struct.pack('>H', len(data))
        self.request.sendall(sz + data)


class DNSServer(UDPServer):
    """Threaded DNS Server (UDP)
    
    Subclass of omniserver.servers.UDPServer that declares self.default_ip and self.record,
    which are needed for resolving incoming queries.Inherits DNS-specific functions from 
    omniserver.servers.BaseDNSServer, general server functions from omniserver.servers.UDPServer
    """
    proto = "DNS"
    def __init__(self, server_address, RequestHandlerClass, default_ip=None):
        super().__init__(server_address, RequestHandlerClass)
        self.default_ip = default_ip
        self.records = []
         
    def add_record(self, zone_record):
        """Add zone-style record to compare incoming queries against
        
        Ex: "google.com 60 IN A 192.168.100.1"
        
        :param zone_record: DNS zone record string to add to server's 'record'
        :type zone_record: str
        """
        self.records.append(*RR.fromZone(zone_record))
        
    def add_zonefile(self, zonefile):
        """Add all DNS records from given DNS zone file
        
        :param zonefile: Path to file containing zone records
        :type zonefile: str
        """
        with open(zonefile) as file:
            rrs = RR.fromZone(file)
            for rr in rrs:
                self.records.append(rr)


class DNSHandler(DNSRequestHandler):
    """Handler class for DNS server requests (UDP)
     
    DNSRequestMixIn overrides incoming(), outgoing(), and response(). 
    response() can be further overriden to customize DNS resolution method. 
    self.records and self.default_ip inherited from DNSBaseHandler
    """        
    def recv(self):
        """Protocol specific method for recieving data to remote socket
        
        Called by recv_data(), which passes the returned data through self.incoming()
        
        :returns data: Data recieved from remote socket
        """
        return self.request[0].strip()
        
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Called by send_data(), after it passes the data through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        self.request[1].sendto(data, self.client_address)


# HTTP #

class HTTPServer(TCPServer):
    """Threaded HTTP server
    
    Subclass of omniserver.servers.TCPServer, works basically the same as http.server.
    TLS/SSL compatible.
    """
    proto = "HTTP"
    server_version = "OmniHTTP/" + __version__
    def __init__(self, server_address, RequestHandlerClass, dir=None):
        super().__init__(server_address, RequestHandlerClass)
        self.working_dir = os.getcwd() if dir is None else dir
        
    def __str__(self):
        return f"{self.proto} Server {self.proto.lower()}://{self.server_address[0]}:{self.server_address[1]}/"
        
    def finish_request(self, request, client_address):
        self.RequestHandlerClass(request, client_address, self, directory=self.working_dir)
        
    def enable_ssl(self, context):
        if super().enable_ssl(context):
            self.proto = "HTTPS"
            return True
        return False


class HTTPHandler(SimpleHTTPRequestHandler):
    """Handler class for HTTP/S server connections/requests
    
    """
    def log_message(self, format, *args):
        """Log arbitrary message

        Overriding BaseHTTPRequestHandler to keep
        all Omniserver logs in the same basic style.
        """
        message = format % args
        print(f"[{self.client_address[0]}:{self.client_address[1]}] {message}")
    
    

""" Server Management
- Classes and functions for starting, stoping, and waiting on threaded servers
- Newly created server objects are appended to the init_servers list,
    so they can be started, stopped, or waited on all at once
    
TODO:
    - Make ServerManager class more comprehensive as a wrapper
        - https://stackoverflow.com/questions/9057669/how-can-i-intercept-calls-to-pythons-magic-methods-in-new-style-classes/9059858#9059858
"""
init_servers = [] # All initialized servers
active_servers = [] # Servers that have been started with start() or start_all()

class ServerManager:
    """Wrapper class for ThreadedServer objects
    
    Contains methods for starting and stopping threads,
    subclasses include more protocol-specific methods.
    """ 
    def __init__(self, ServerClass):
        self.server = ServerClass
        init_servers.append(self)
        self.thread = None
        
    def __str__(self):
        return str(self.server)
        
    def __getattr__(self, attr):
        """Provides proxy access to the server objects functions/attributes
        """
        return getattr(self.server, attr)
    
    def start(self, daemon=True):
        """Create and start server process thread (non-blocking)
        
        :param daemon: Daemonize thread
        :type daemon: bool
        """
        if self.thread is not None: # Do nothing if thread has already been started
            return
        self.server.activate()
        active_servers.append(self)
        
        self.thread = threading.Thread(target=self.server.serve_forever, daemon=daemon)
        self.thread.start() # Thread object available for whatever
        return self # Returning self so wait() method can be used on same line
            
    def wait(self, interrupt=KeyboardInterrupt):
        """Wait (block) for server thread, shutdown upon interrupt
        
        :param interrupt: Exception to trigger the shutdown of active thread
        """
        if self.thread is None: # Do nothing if no thread has been started
            return
        print("(Ctrl-C to exit)\n")
        try:
            while 1:
                sleep(1)
                sys.stderr.flush()
                sys.stdout.flush()
        except interrupt:
            pass
        finally:
            self.stop()
            
    def stop(self):
        """Shutdown/close server
        """
        self.server.shutdown()
        active_servers.remove(self)
        self.thread = None


"""Functions for controlling all active servers
"""

def stop_all():
    """Shutdown all currently active server threads
    """
    if len(active_servers) > 0:
        for server in active_servers:
            server.stop()

def wait_all(interrupt=KeyboardInterrupt): # Might replace with signals
    """ Wait indefinetly to allow all non-blocking threads to operate
    
    Shutdown all servers upon triggering interrupt
    
    :param interrupt: Exception to trigger the shutdown of all active threads
    """
    print("(Ctrl-C to exit)\n")
    if len(active_servers) <= 0: # Don't wait if no servers are started
        return
    try:
        while 1:
            sleep(1)
            sys.stderr.flush()
            sys.stdout.flush()
    except interrupt:
        pass
    finally:
        stop_all()

def start_all(wait=True):
    """Start all currently initialized server threads
    
    :param wait: Wait (block) until interrupt is triggered, then shutdown all active servers
    :type wait: bool
    """
    if len(init_servers) > 0:
        for server in init_servers:
            server.start()
        if wait:
            wait_all()
            
   


#
# Copyright (C) 2023 LLCZ00
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.  
#
"""
omniserver.core:
    - Main module functions for initializing client and server objects

TODO:
    - 
"""
from omniserver import servers, clients


""" Server Initialization Functions
"""

def tcp_server(port=0, ip='', *, RequestHandler=servers.TCPHandler, ServerHandler=servers.TCPServer, sslcontext=None):
    """Initialize TCPServer and return its manager object
    
    :param port: Port to listen on (Default: random high port)
    :type port: int
    :param ip: IP address to accept connections with (Default: 0.0.0.0)
    :type ip: str
    :param RequestHandler: servers.TCPHandler class or subclass to process requests with
    :param ServerHandler: servers.TCPServer class or subclass to handle server socket
    :param sslcontext: SSLContext to wrap socket with
    :type context: ssl.SSLContext

    :returns ServerManager: ServerManager object wrapped around TCPServer
    """        
    server = servers.ServerManager(ServerHandler((ip, port), RequestHandler))    
    if sslcontext:
        server.enable_ssl(context)
    return server


def udp_server(port=0, ip='', *, RequestHandler=servers.UDPHandler, ServerHandler=servers.UDPServer):
    """Initialize UDPServer and return its manager object
    
    :param port: Port to listen on (Default: random high port)
    :type port: int
    :param ip: IP address to accept connections with (Default: 0.0.0.0)
    :type ip: str
    :param RequestHandler: servers.UDPHandler class or subclass to process requests with
    :param ServerHandler: servers.UDPServer class or subclass to handle server socket
    
    :returns ServerManager: ServerManager object wrapped around UDPServer
    """
    return servers.ServerManager(ServerHandler((ip, port), RequestHandler))


def dns_tcp_server(port=53, ip='', *, RequestHandler=servers.DNSHandlerTCP, ServerHandler=servers.DNSServerTCP, default_ip=None, record=None, zonefile=None, sslcontext=None):
    """Initialize DNSServerTCP and return its manager object
    
    :param port: Port to listen on (Default: 53)
    :type port: int
    :param ip: IP address to accept connections with (Default: 0.0.0.0)
    :type ip: str
    :param RequestHandler: servers.DNSHandlerTCP class or subclass to process requests with
    :param ServerHandler: servers.DNSServerTCP class or subclass to handle server socket
    :param default_ip: Default IP address to answer DNS queries with, if not resolved by other means
    :type default_ip: str
    :param record: Zone-style DNS record to compare queries against
    :type record: str
    :param zonefile: Zone file of DNS records to populate server's records with
    :type zonefile: str
    :param sslcontext: SSLContext to wrap socket with
    :type sslcontext: ssl.SSLContext
    
    :returns ServerManager: ServerManager object wrapped around DNSServerTCP
    """        
    server = servers.ServerManager(ServerHandler((ip, port), RequestHandler, default_ip=default_ip))
    if zonefile:
        server.add_zonefile(zonefile)
    if record:
        server.add_record(record)
    if sslcontext:
        server.enable_ssl(context)
    return server


def dns_server(port=53, ip='', *, RequestHandler=servers.DNSHandler, ServerHandler=servers.DNSServer, default_ip=None, record=None, zonefile=None):
    """Initialize DNSServerTCP and return its manager object
    
    :param port: Port to listen on (Default: 53)
    :type port: int
    :param ip: IP address to accept connections with (Default: 0.0.0.0)
    :type ip: str
    :param RequestHandler: servers.DNSHandler class or subclass to process requests with
    :param ServerHandler: servers.DNSServer class or subclass to handle server socket
    :param default_ip: Default IP address to answer DNS queries with, if not resolved by other means
    :type default_ip: str
    :param record: Zone-style DNS record to compare queries against
    :type record: str
    :param zonefile: Zone file of DNS records to populate server's records with
    :type zonefile: str
        
    :returns ServerManager: ServerManager object wrapped around DNSServer
    """        
    server = servers.ServerManager(ServerHandler((ip, port), RequestHandler, default_ip=default_ip))
    if zonefile:
        server.add_zonefile(zonefile)
    if record:
        server.add_record(record)
    return server    
 
 
def http_server(port=8080, ip='', *, RequestHandler=servers.HTTPHandler, ServerHandler=servers.HTTPServer, dir=None, sslcontext=None):
    """Initialize HTTPServer and return its manager object
    
    :param port: Port to listen on (Default: 8080)
    :type port: int
    :param ip: IP address to accept connections with (Default: 0.0.0.0)
    :type ip: str
    :param RequestHandler: servers.HTTPHandler class or subclass to process requests with
    :param ServerHandler: servers.HTTPServer class or subclass to handle server socket
    :param dir: Server's working directory (Default: CWD)
    :type dir: str
    :param sslcontext: SSLContext to wrap socket with
    :type context: ssl.SSLContext    
    :returns ServerManager: ServerManager object wrapped around HTTPHandler
    """
    server = servers.ServerManager(ServerHandler((ip, port), RequestHandler, dir))
    if sslcontext:
        server.enable_ssl(context)
    return server   



"""Client Initialization Functions
"""

def tcp_client(remote_addr: tuple, *, Handler=clients.TCPClient, sslcontext=None, hostname=None):
    """Initialize and return clients.TCPClient object
    
    :param remote_addr: IP address (or hostname) and port of remote socket
    :type remote_addr: tuple
    :param Handler: clients.TCPClient class or subclass to handle data exchange
    :type Handler: clients.TCPClient
    :param sslcontext: SSLContext to wrap socket with (If None, create context using **kwargs)
    :type sslcontext: ssl.SSLContext
    :param hostname: Hostname of remote server (Required for SSL, unless hostname defined in remote_addr)
    :type hostname: str       
    :returns TCPClient: Client object for exchanging data with remote TCP server
    """
    client = Handler(remote_addr)
    if sslcontext:
        client.enable_ssl(sslcontext, hostname)
    return client


def udp_client(remote_addr, *, Handler=clients.UDPClient):
    """Initialize and return clients.UDPClient object
    
    :param remote_addr: IP address (or hostname) and port of remote socket
    :type remote_addr: tuple
    :param Handler: clients.UDPClient class or subclass to handle data exchange
    :type Handler: clients.UDPClient    
    :returns UDPClient: Client object for exchanging data with remote UDP server
    """
    return Handler(remote_addr)


def dns_client(remote_addr, *, Handler=clients.DNSClient):
    """Initialize and return clients.DNSClient object (UDP)
    
    :param remote_addr: IP address (or hostname) and port of remote socket
    :type remote_addr: tuple
    :param Handler: clients.DNSClient class or subclass to handle data exchange
    :type Handler: clients.DNSClient   
    :returns DNSClient: Client object for sending queries to remote DNS server (UDP)
    """
    return Handler(remote_addr)


def dns_tcp_client(remote_addr, *, Handler=clients.DNSClientTCP, sslcontext=None):
    """Initialize and return clients.DNSClientTCP object (TCP)
    
    :param remote_addr: IP address (or hostname) and port of remote socket
    :type remote_addr: tuple
    :param Handler: clients.DNSClientTCP class or subclass to handle data exchange
    :type Handler: clients.DNSClientTCP
    :param sslcontext: SSLContext to wrap socket with (If None, create context using **kwargs)
    :type sslcontext: ssl.SSLContext    
    :returns DNSClientTCP: Client object for sending queries to remote DNS server (TCP)
    """
    client = Handler(remote_addr)
    if sslcontext:
        client.enable_ssl(sslcontext, hostname)
    return client


def http_client(remote_addr, *, Handler=clients.HTTPClient, dir=None, sslcontext=None, hostname=None):
    """Initialize and return clients.HTTPClient object
    
    :param remote_addr: IP address (or hostname) and port of remote socket
    :type remote_addr: tuple
    :param Handler: clients.HTTPClient class or subclass to handle data exchange
    :type Handler: clients.HTTPClient
    :param dir: Working directory (Default: CWD)
    :type dir: str
    :param sslcontext: SSLContext to wrap socket with (If None, create context using **kwargs)
    :type sslcontext: ssl.SSLContext
    :param hostname: Hostname of remote server (Required for SSL, unless hostname defined in remote_addr)
    :type hostname: str
    :returns HTTPClient: Client object for exchanging data with remote web server (via GET and POST requests)
    """
    client = Handler(remote_addr, dir)
    if sslcontext:
        client.enable_ssl(context, hostname)
    return client









    





#
# Copyright (C) 2023 LLCZ00
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.  
#
"""
omniserver.clients:
    - Classes for client objects

Main inheritable classes:
    TCPClient
    UDPClient
    DNSClient
    DNSClientTCP
    HTTPClient
    
TODO:
"""   
import socket
import http.client
import struct
import os
from time import sleep
from dnslib import DNSRecord

from omniserver import certs

__all__ = ["TCPClient", "UDPClient", "DNSClientTCP", 
            "DNSClient", "HTTPClient"]


"""ClientHandler classes
- Inherited by ClientHandler subclasses in omniserver.clients to provide overridable data handling methods
"""

class BaseClientHandler:
    """Base class for TCP and UDP Client subclasses

    Contains overridable methods for finer control over socket, data, and client actions.
    Subclasses must override send() and recv() with their specific implimentations.
    The send/recv and send_data/recv_data methods are seperated so as to easily 
    allow for bypassing the incoming/outgoing filter methods, if necessary
    """
    verbose = True
    def __init__(self, remote_addr: tuple, socket=None):
        self.remote_addr = remote_addr
        self.sock = socket     
        if self.sock is None:
            self.sock_setup()
            
    def __enter__(self):
        return self
        
    def __exit__(self, type, val, traceback):
        try:
            self.close()
        except:
            pass
            
    def sock_setup(self):
        """Overridable method to initialize protocol-specific socket
        
        Required
        """
        raise NotImplementedError
            
    def incoming(self, data):
        """Process incoming data as it's recieved from self.recv()
        """
        try:
            data = data.decode()
        except:
            pass
        return data.strip()
        
    def outgoing(self, resp):
        """Process outgoing response/data before it's sent via self.send()
        """
        if type(resp) is not bytes:
            resp = f"{resp}\n".encode("utf-8")
        return resp
        
    def close(self):
        try:
            if self.sock:
                self.sock.close()
        except:
            pass
 

"""Public ClientHandler subclasses
"""
class TCPClient(BaseClientHandler):
    """Handler class for TCP client objects
    
    Used for exchanging data with remote TCP socket.
    TLS/SSL compatible.
    """
    def sock_setup(self):
        """Initialize TCP socket and 'connected flag'
        """
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.connected = False
        
    def __enter__(self):
        self.connect()
        return self
        
    def recv(self, buffer=1024):
        """Protocol specific method for recieving data from remote socket
        
        :param buffer: Size of recieve buffer
        :type buffer: int
        :returns data: Data recieved, after being processed through self.incoming()
        """
        if not self.connected:
            raise Exception("Not connected to remote server")
        data = self.incoming(self.sock.recv(buffer))
        if data and self.verbose:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] Data recieved: {data}")
        return data
        
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        :param data: Data to be sent to remote socket, after it's been passed through self.outgoing()
        """
        if not self.connected:
            raise Exception("Not connected to remote server")
        self.sock.sendall(self.outgoing(data))
        if self.verbose:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] Data sent.")
        
    def connect(self):
        """Attempt to connect to remote address
        
        :rtype: bool
        """
        if self.connected:
            return True
        try:
            self.sock.connect(self.remote_addr)
        except:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] Failed to establish TCP connection.")
            return False
        else:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] TCP connection established.")
            self.connected = True
            return True
        
    def enable_ssl(self, context=None, server_hostname=None):
        """Enable TLS/SSL on client socket
        
        :param context: SSLContext to wrap socket with
        :type context: SSLContext
        :param server_hostname: Hostname of remote server
        :type server_hostname: str
        
        :rtype: bool
        """
        if context is None:
            context = certs.create_client_context() # default context
        if server_hostname is None:
            server_hostname = self.remote_addr[0]
        try:
            self.sock = context.wrap_socket(self.sock, server_side=False, server_hostname=server_hostname)
        except Exception as e:
            print("TLS/SSL failed.")
            raise e    
        return True
            
    def beacon(self, freq=5, tries=0, timeout=None):
        """Periodically attempt to connect to remote socket
        
        :param freq: Frequency, in seconds, to attempt connection
        :type freq: float
        :param tries: Total number of connection attempts before returning False (0=infinity)
        :type param: int
        :param timeout: Set socket timeout time, in seconds
        
        :rtype: bool
        """
        self.sock.settimeout(timeout)
        attempt = 1
        while attempt != tries:
            if self.connect():
                return True
            attempt += 1
            sleep(freq)
        return False
        
    def close(self):
        """Shutdown and close socket
        """
        try:
            if self.sock:
                self.sock.shutdown(1)
                self.sock.close()
                self.connected = False
        except:
            pass


class UDPClient(BaseClientHandler):
    """Handler class for UDP client objects
    
    Used to exchange data with remote UDP server.
    """
    def sock_setup(self):
        """initialize UDP socket
        """
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
    def recv(self, buffer=1024):
        """Protocol specific method for recieving data from remote socket
        
        :param buffer: Size of recieve buffer
        :type buffer: int
        :returns data: Data recieved, after being processed through self.incoming()
        """
        try:
            data = self.incoming(self.sock.recv(buffer))
        except:
            print("Unable to reach remote server")
            return None
        if data and self.verbose:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] UDP data recieved: {data}")
        return data
        
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Sends data after it's been passed through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        self.sock.sendto(self.outgoing(data), self.remote_addr)
        if self.verbose:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] Data sent.")
        
    def beacon(self, msg="Beacon message", freq=5, tries=0, timeout=None):
        """Periodically attempt to connect to remote socket
        
        Returns True upon recieving UDP data back from remote socket.
        
        :param msg: Default data/message to send to remote socket
        :type msg: str
        :param freq: Frequency, in seconds, to attempt connection
        :type freq: float
        :param tries: Total number of connection attempts before returning False (0=infinity)
        :type param: int
        :param timeout: Set socket timeout time, in seconds
        
        :rtype: bool
        """
        self.sock.settimeout(timeout)
        attempt = 1
        while attempt != tries:
            self.send(msg)
            if self.recv():
                return True
            attempt += 1
            sleep(freq)
        return False


class DNSClientMixIn:
    """MixIn class to extend UDPClient or TCPClient with DNS-specific methods
    
    Overrides incoming() and outgoing() methods, adds query() method
    """
    verbose = True
    def incoming(self, data):
        """
        Process incoming data as it's recieved from remote server
        """
        try:
            response = str(DNSRecord.parse(data.strip()).get_a().rdata)
        except:
            raise Exception("Error parsing DNS response")
        return response
        
    def outgoing(self, query):
        """
        Process outgoing response/data before it's sent to remote server
        """
        if type(query) is not bytearray:
            try:
                query = query.pack()
            except:
                raise Exception("Error packing DNS query")
        return query
        
    def query(self, domain, record_type="A"):
        """Send given query to remote server
        
        :param domain: Domain name to query
        :type domain: str
        :param record_type: Record type to query for
        :type record_type: str
        
        :returns answer: IP address returned from DNS server
        :rtype: str
        """
        query = DNSRecord.question(domain, qtype=record_type)
        self.send(query)
        response = self.recv()
        if response and self.verbose:
            print(f"[{self.remote_addr[0]}:{self.remote_addr[1]}] DNS response: {response}")
        
        
class DNSClientTCP(DNSClientMixIn, TCPClient):
    """
    Handler Class for DNS client objects (TCP)
    
    Overrides serveral BaseClientHandler methods to provide DNS-specific 
    functionality by inheriting DNSClientMixIn.
    Inherits socket-specific functionality from TCPClient
    """
    def recv(self, buffer: int = 8192):
        """
        Main method to be used for recieving data
        - Returns data from recv(), after being passed through incoming()
        """
        if not self.connected:
            raise Exception("Not connected to remote server")
        data = self.sock.recv(buffer).strip()
        sz = struct.unpack('>H', data[:2])[0]
        if sz < len(data) - 2:
            raise Exception("Wrong size of TCP packet")
        elif sz > len(data) - 2:
            raise Exception("Too big TCP packet")
        return self.incoming(data[2:])
        
    def send(self, resp):
        """
        Main method to be used for sending data
        - Sends data after it's passed through outgoing()
        """
        if not self.connected:
            raise Exception("Not connected to remote server")
        data = self.outgoing(resp)
        sz = struct.pack('>H', len(data))
        self.sock.sendall(sz + data)


class DNSClient(DNSClientMixIn, UDPClient):
    """
    Handler Class for DNS client objects (UDP)
    
    Overrides serveral BaseClientHandler methods to provide DNS-specific 
    functionality by inheriting DNSClientMixIn.
    Inherits socket-specific functionality from UDPClient
    """
    def recv(self, buffer=1024):
        """Protocol specific method for recieving data from remote socket
        
        :param buffer: Size of recieve buffer
        :type buffer: int
        :returns data: Data recieved, after being processed through self.incoming()
        """
        try:
            data = self.incoming(self.sock.recv(buffer))
        except:
            print("Unable to reach remote server")
            return None
        return data
        
    def send(self, data):
        """Protocol specific method for sending data to remote socket
        
        Sends data after it's been passed through self.outgoing()
        
        :param data: Data to be sent to remote socket
        """
        self.sock.sendto(self.outgoing(data), self.remote_addr)


class HTTPClient(http.client.HTTPConnection):
    """Handler Class for HTTP client objects
    
    Used for issuing requests to remote HTTP/S server.
    Subclass of http.client.HTTPConnection
    """
    proto = "HTTP"
    verbose = True
    def __init__(self, remote_addr, dir=None):
        super().__init__(host=remote_addr[0], port=remote_addr[1], timeout=socket._GLOBAL_DEFAULT_TIMEOUT, source_address=None, blocksize=8192)
        self.dir = dir if dir else os.getcwd()
        self.tls = False
        self.context = None
        self.hostname = None
        
    def __enter__(self):
        self.connect()
        return self
        
    def __exit__(self, type, val, traceback):
        pass # HTTPConnection closes automatically anyway
        
    def __str__(self):
        return f"[{self.host}:{self.port}] {self.proto}"
        
    def incoming(self, url, response):
        """Process response data recieved from get()
        
        Default: Write data to file
        """
        if response.getcode() == 200:
            with open(os.path.join(self.dir, os.path.basename(url)), "wb") as file:
                file.write(response.read())
        
    def get(self, url, body=None, headers={}, *, encode_chunked=False):
        """Send GET request to remote server
        
        :param url: URL of resource to request
        :type url: str
        :param body: Data to be sent after headers are finished
        :param headers: HTTP headers
        :type headers: dict
        :param encode_chunked: Chunk-encode body
        :type encode_chunked: bool
        """
        self._send_request("GET", url, body, headers, encode_chunked) # http.client.HTTPConnection
        resp = self.getresponse()
        if self.verbose:
            print(f"{self} request: 'GET {url}' -> {resp.getcode()}")
        self.incoming(url, resp)
        return resp.getcode()
        
    def post(self, url, body=None, headers={}, *, encode_chunked=False):
        """Send POST request to remote server
        
        :param url: URL of resource to request
        :type url: str
        :param body: Data to be sent after headers are finished
        :param headers: HTTP headers
        :type headers: dict
        :param encode_chunked: Chunk-encode body
        :type encode_chunked: bool
        """
        self._send_request("POST", url, body, headers, encode_chunked) # http.client.HTTPConnection
        resp = self.getresponse()
        if self.verbose:
            print(f"{self} request: 'POST {url}' -> {resp.getcode()}")
        return resp.getcode()
        
    def connect(self):
        """Establish connection to remote server
        """
        if self.sock is None:
            super().connect()
            if self.tls:
                self.enable_ssl()    

    def enable_ssl(self, context=None, server_hostname=None):
        """Enable TLS/SSL on client socket
        
        :param context: SSLContext to wrap socket with
        :type context: SSLContext
        :param server_hostname: Hostname of remote server
        :type server_hostname: str
        
        :rtype: bool
        """
        self.tls = True
        if context is None:
            if self.context is None:
                context = certs.create_client_context() # Default context
            else:
                context = self.context
            
        if server_hostname:
            self.hostname = server_hostname
        elif self.hostname is None:
            self.hostname = self.host
            
        if self._http_vsn == 11:
            context.set_alpn_protocols(['http/1.1'])
        if context.post_handshake_auth is not None:
            context.post_handshake_auth = True
            
        if self.sock is None: # Assign self.context and return if socket hasn't been declared yet
            self.context = context
            return False

        try:
            self.sock = context.wrap_socket(self.sock, server_side=False, server_hostname=self.hostname)
        except Exception as e:
            print("TLS/SSL failed.")
            raise e
        
        self.context = context
        self.proto = "HTTPS"
        return True
    
    



import os
from setuptools import setup, find_packages


with open(os.path.join("omniserver", "version.py"), "r") as vfile:
    exec(vfile.read())


this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.md"), "r") as file:
    readme = file.read()

requirements = [
    "dnslib"
    ]

setup(
    name="omniserver",
    version=__version__,
    author="LLCZ00",
    description="Module for network testing and prototyping",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/LLCZ00/Omniserver",
    license="Apache 2.0",
    keywords="malware analysis networking",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Security",
        "Topic :: Utilities",
    ],
    python_requires=">=3.8",
    zip_safe=False,
    install_requires=requirements,
    packages=find_packages()
)




#
# Copyright (C) 2023 LLCZ00
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.  
#
"""
omniserver.utilities:
    - SSL wrappers/handles

Command to generate self-signed TLS certificate & private key:
openssl req -new -newkey rsa:4096 -x509 -sha256 -days 365 -nodes -out cert.crt -keyout private.key

Server certs signed by CA reference: https://gist.github.com/toolness/3073310

TODO:
    - Add CSR stuff
"""
import subprocess
import os
import sys
import ssl


__all__ = ["create_client_context", "create_server_context",
            "update_ca"]
if sys.platform == "linux":
    __all__.extend(["create_cert_key", "read_cert"])

"""
TLS/SSL Functions
"""
def create_client_context(protocol=ssl.PROTOCOL_TLS_CLIENT,
        cert_required=True,
        ca_cert=None,
        ca_path=None,
        ca_data=None,
        certfile=None,
        keyfile=None):
        """Create SSL context for client sockets
        
        Arguments can be passed to initialization functions as **kwargs,
        or this function can be called directly.
        
        :param protocol: Protocol to be passed to ssl.SSLContext()
        :param cert_required: Require a valid certificate from the server
        :type cert_required: bool
        :param ca_cert: Path to certificate to be used to verify server against
        :type ca_cert: str
        :param ca_path: Path to folder with certificates to verify server against
        :type ca_path: str
        :param ca_data: Ascii string of PEM-encoded cert, or DER-encoded bytes
        :param certfile: Certificate to be sent if asked to verify self
        :type certfile: str
        :param keyfile: Private key file
        :type keyfile: str
        
        :returns context: SSLContext that can be used to wrap a client socket
        :rtype: ssl.SSLContext
        """      
        ctx = ssl.SSLContext(protocol)
        if cert_required:
            ctx.check_hostname = True
            ctx.verify_mode = ssl.CERT_REQUIRED
            if ca_cert or ca_path or ca_data:
                try:
                    ctx.load_verify_locations(ca_cert, ca_path, ca_data)
                except Exception as e:
                    print("Exception occured loading CA certs")
                    print(e)
            else:
                ctx.load_default_certs()
        else: 
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        if certfile:
            ctx.load_cert_chain(certfile, keyfile)     
        return ctx
        
def create_server_context(protocol=ssl.PROTOCOL_TLS_SERVER, certfile=None, keyfile=None):
        """Create SSL context for server sockets
        
        Arguments can be passed to initialization functions as **kwargs,
        or this function can be called directly.
        
        :param protocol: Protocol to be passed to ssl.SSLContext()
        :param certfile: Certificate to be sent if asked to verify self
        :type certfile: str
        :param keyfile: Private key file (Not required if in certfile)
        :type keyfile: str
        
        :returns context: SSLContext that can be used to wrap a server socket
        :rtype: ssl.SSLContext
        """
        ctx = ssl.SSLContext(protocol)
        ctx.load_cert_chain(certfile, keyfile)
        return ctx

def create_cert_key(cert="ca.crt", key="private.key", keysize=4096, days=365, **kwargs):
    """Generate self signed cert and private key
    
    Create self signed CA cert and private key, primarily for use by servers.
    Currently only available on Linux.
    
    :param cert: Name of created cert
    :type cert: str
    :param key: Name of created private key
    :type key: str
    :param keysize: Bit-size of private key encryption
    :type keysize: int
    :param days: Days
    :type days: int
    :param **kwargs: Cert subj information (C, ST, L, O, OU, CN)
    
    :returns cert_key: Absolute path to cert and private key
    :rtype: tuple
    """
    if sys.platform == "win32":
        raise NotImplementedError

    cmd = f"openssl req -new -newkey rsa:{str(keysize)} -x509 -sha256 -days {str(days)} -nodes -out {cert} -keyout {key}".split(" ")
    subj_cmd = ""

    subj = {
        "C":"US",
        "ST":"ND",
        "L":"Fargo",
        "O":"",
        "OU":"",
        "CN":"placeholder.com"
    }
    # Probably a better way to do this
    for arg in kwargs:
        if arg in subj:
            subj[arg] = kwargs[arg]
    
    for topic in subj:
        if subj[topic]:
            subj_cmd += f"/{topic}={subj[topic]}"

    if subj_cmd:
        cmd.append("-subj")
        cmd.append(subj_cmd)

    proc = subprocess.run(cmd)
    if proc.returncode != 0:
        print("Error generating cert/private key")
        return None
        
    return os.path.abspath(cert), os.path.abspath(key)


def update_ca(certfile: str, remove: bool = False):
    """Add or remove root certificate to list of trusted certs 
    
    :param certfile: Path to CA file
    :type certfile: str
    :param remove: Remove given cert from trust CA list
    :type remove: bool
    """
    if not remove and not os.path.exists(certfile):
        raise Exception(f"Certificate not found '{certfile}'")
    
    if sys.platform == "linux":
        if remove:
            subprocess.run(["sudo", "update-ca-certificates", "--fresh"], check=True)
        else:
            subprocess.run(["sudo", "cp", certfile, "/usr/local/share/ca-certificates/"], check=True)
            subprocess.run(["sudo", "update-ca-certificates"], check=True)
        
    elif sys.platform == "win32":
        if remove:
            certid = "0x15" # Run command to get certid for certfile
            subprocess.run(['certutil', '-delstore', '"ROOT"', certid], check=True)
        else:
            subprocess.run(['certutil', '-addstore', '-f', '"ROOT"', certfile], check=True)
    else:
        print(f"Failed to install CA, unsupported platform '{sys.platform}'")


def read_cert(certpath):
    """Print certificate information
    
    Linux only
    """
    if sys.platform == "win32":
        raise NotImplementedError
    subprocess.run(f"openssl x509 -in {certpath} -text -noout".split(" "))

